from pytorch_dragon.pytorch_dragon import sobel_filter
import cv2 as cv
import numpy as np
from timeit import default_timer as timer


class TestSobel:
    def test_sobel(self):
        path_to_ims = "./test/images"
        # CPU block
        im1 = cv.imread(f"{path_to_ims}/image1.jpg")
        # convert to grayscale
        gim1 = cv.cvtColor(im1, cv.COLOR_BGR2GRAY)
        im2 = np.zeros_like(im1)
        im2[:, :, 0] = (
            gim1  # To display the image correctly you'll need to send the grey channel to each of the three image channels
        )
        im2[:, :, 1] = gim1
        im2[:, :, 2] = gim1

        # cuda
        print("Starting CUDA Sobel Filter calc")

        start = timer()
        output = sobel_filter(gim1)
        end = timer()
        print(f"CUDA Sobel Filter calculation took: {round(end-start, 2)}s")
        assert end - start < 10.0



import pytest
from backprop.grad_accumulator import GradAccumulator
import torch as T
import logging
import numpy as np

initial_data = T.tensor(
    data=[1.0, 1.1, 1.21],
    requires_grad=False,
)

test_data = T.tensor(
    data=[1.51, 1.6, 1.1, 1.3, 0.9, 0.8, 0.8, 1.1, 0.4, 0.5],
    requires_grad=False,
)

test_data_2 = T.tensor(
    data=[1.0, 1.1, 1.21, 1.51, 1.6, 1.1, 1.3, 0.9, 0.8, 0.8, 1.1, 0.4, 0.5],
    requires_grad=False,
)


class Test_Package:
    # test accumulator init
    def test_accum_init(self):
        try:
            accum = GradAccumulator(K=5, start_data=initial_data, requires_grad=False)
        except:
            assert False

        assert True

    # test accumulator window length bound
    def test_accum_length_bound(self):
        accum = GradAccumulator(K=5, start_data=initial_data, requires_grad=False)
        for i in range(10):
            item = test_data[i]
            output = accum(item)
            if len(accum.window.data) > accum.K:
                assert False
        assert True

    # test accumulator window length bound
    def test_accum_iter_accuracy(self):
        accum = GradAccumulator(K=3, start_data=initial_data, requires_grad=False)
        first_3_vals = [1.2733, 1.4400, 1.4033]  # ground truth values for test
        correct = 0
        for i in range(3):
            item = test_data[i]
            output = accum(item)
            if output.item() - first_3_vals[i] <= 0.0001:
                correct += 1
        if correct == 3:
            assert True
        else:
            assert False

    def test_accum_in_training_run(self):
        accum = GradAccumulator(K=3, requires_grad=False)
        ground_truth = [
            1.0,
            1.05,
            1.1033,
            1.2733,
            1.4400,
            1.4033,
            1.3333,
            1.1000,
            1.0000,
            0.8333,
            0.9000,
            0.7667,
            0.6667,
        ]  # all have been hand calculated
        for i in range(len(test_data_2)):
            item = test_data_2[i]
            output: T.tensor = accum(item)
            assert (output.item() - ground_truth[i]) <= 0.0001

    def test_accum_not_modify_tensor_params(self):
        accum = GradAccumulator(K=3, start_data=initial_data, requires_grad=False)
        for i in range(len(test_data_2)):
            item = test_data_2[i]
            output: T.tensor = accum(item)
            assert item.requires_grad == output.requires_grad



import torch as T
import numpy as np
from torch.distributions.normal import Normal
import torch.nn.functional as F
from botorch.acquisition import ProbabilityOfImprovement, ExpectedImprovement
from gpytorch.likelihoods.gaussian_likelihood import GaussianLikelihood

from search.acquisition.functions import probability_improvement, expected_improvement
from search.models.bopt_models import DragonGPR
from search.continuous_search import BaseDragonBayesOpt
from search.hyperparam import Hyperparameter, Constraint, RangeConstraint
from typing import Optional

from test_utils.test_models import LeNet, LSTMTS, LSTMTSConfig, device
import pytest

# TODO: Write and test bayes opt call / fit / predict functions


class TestBayesOpt:
    def test_bopt_iteration_stop(self):
        # initialize hyperparameters (test fn: lr cannot be a float % 2 == 0)
        T.manual_seed(0)
        np.random.seed(0)

        def lr_constraint(value: float) -> bool:
            return int(100 * value) % 2 == 0

        lr_param = Hyperparameter(
            name="lr",
            type_="numerical",
            x=0.01,
            constraints=[Constraint(lr_constraint)],
            range_=(1e-10, 1e-2),
        )

        beta0_param = Hyperparameter(
            name="beta0",
            type_="numerical",
            x=0.91,
            range_=(0.9, 0.95),
        )

        beta1_param = Hyperparameter(
            name="beta1",
            type_="numerical",
            x=0.97,
            range_=(0.95, 0.999999),
        )
        print("Finished lr param init")

        def bad_model(x, lr):
            return 10 * x[0][0]

        # objective function
        def obj(model: callable, input_):
            return model(input_)

        bopt = BaseDragonBayesOpt(
            objective_function=obj,
            Y_init=T.tensor([1.0]),
            hyper_params=[lr_param, beta0_param, beta1_param],
            iters=0,
        )

        result = bopt(model=bad_model, batch_X=T.tensor([1]), batch_Y=T.tensor([0.0]))
        assert result["Y"] == T.tensor([1.0])
        assert bopt.stop

    def test_two_iters_lstm(self):
        # seed
        T.manual_seed(0)
        np.random.seed(0)

        # model init
        model_config = LSTMTSConfig()
        model = LSTMTS(model_config)

        lr_param = Hyperparameter(
            name="lr", type_="numerical", x=0.01, range_=(1e-10, 1e-2)
        )

        beta0_param = Hyperparameter(
            name="beta0", type_="numerical", x=0.91, range_=(0.9, 0.95)
        )

        beta1_param = Hyperparameter(
            name="beta1", type_="numerical", x=0.97, range_=(0.95, 0.999999)
        )

        x_train = T.tensor(
            [
                [0.1, 0.2, 0.3, 0.4, 0.5, 0.6],
                [0.05, 0.25, 0.3, 0.4, 0.5, 0.6],
                [0.1, 0.1, 0.3, 0.3, 0.6, 0.6],
            ],
        )
        y_train = T.tensor([[1.0], [1.0], [0.0]])

        def obj(model, X, y, hprev, cprev, prediction_name: Optional[str] = "call"):
            count = 0.0
            for i in range(X.shape[0]):
                x = X[i]
                (y, _) = model(X, hprev=hprev, cprev=cprev)
                count += T.sum(y)
            return count

        bopt = BaseDragonBayesOpt(
            objective_function=obj,
            Y_init=T.tensor([0.5]),
            hyper_params=[lr_param, beta0_param, beta1_param],
            iters=50,
        )

        result1 = bopt(
            model,
            x_train,
            y_train,
            hprev=T.zeros(1, 1, 24),
            cprev=T.zeros(1, 1, 24),
        )

        result2 = bopt(
            model,
            x_train,
            y_train,
            hprev=T.zeros(1, 1, 24),
            cprev=T.zeros(1, 1, 24),
        )

        # check improvement + different results
        assert not (-1 * result1["value"] == -1 * result2["value"])

        # check all hyperparams have their values set.
        returned_vals = result2["x"]
        for i in range(len(result2["x"])):
            assert returned_vals[i] == bopt.params[i].value

    def test_banditos(self):
        # seed
        T.manual_seed(0)
        np.random.seed(0)

        # model init
        model_config = LSTMTSConfig()
        model = LSTMTS(model_config)

        lr_param = Hyperparameter(
            name="lr", type_="numerical", x=0.01, range_=(1e-10, 1e-2), log_range=True
        )

        beta0_param = Hyperparameter(
            name="beta0", type_="numerical", x=0.91, range_=(0.9, 0.95), log_range=True
        )

        beta1_param = Hyperparameter(
            name="beta1",
            type_="numerical",
            x=0.97,
            range_=(0.95, 0.999999),
            log_range=True,
        )

        x_train = T.tensor(
            [
                [0.1, 0.2, 0.3, 0.4, 0.5, 0.6],
                [0.05, 0.25, 0.3, 0.4, 0.5, 0.6],
                [0.1, 0.1, 0.3, 0.3, 0.6, 0.6],
            ],
        )
        y_train = T.tensor([[1.0], [1.0], [0.0]])

        def obj(model, X, y, hprev, cprev, prediction_name: Optional[str] = "call"):
            count = 0.0
            for i in range(X.shape[0]):
                x = X[i]
                (y, _) = model(X, hprev=hprev, cprev=cprev)
                count += T.sum(y)
            return count

        bopt = BaseDragonBayesOpt(
            objective_function=obj,
            Y_init=T.tensor([0.5]),
            hyper_params=[lr_param, beta0_param, beta1_param],
            iters=50,
            regressor_type="banditos",
        )

        result1 = bopt(
            model,
            x_train,
            y_train,
            hprev=T.zeros(1, 1, 24),
            cprev=T.zeros(1, 1, 24),
        )

        result2 = bopt(
            model,
            x_train,
            y_train,
            hprev=T.zeros(1, 1, 24),
            cprev=T.zeros(1, 1, 24),
        )

        # check improvement + different results
        assert not (-1 * result1["value"] == -1 * result2["value"])

        # check all hyperparams have their values set.



import torch as T
import numpy as np
from torch.distributions.normal import Normal
import torch.nn.functional as F
from botorch.acquisition import ProbabilityOfImprovement, ExpectedImprovement
from gpytorch.likelihoods.gaussian_likelihood import GaussianLikelihood

from search.acquisition.functions import probability_improvement, expected_improvement
from search.models.bopt_models import DragonGPR
from search.continuous_search import BaseDragonBayesOpt
from search.hyperparam import Hyperparameter, Constraint, RangeConstraint

from test_utils.test_models import LeNet, LSTMTS, LSTMTSConfig, device

X = T.normal(
    mean=T.zeros(size=(3, 3)),
    std=1,
)
Y = T.tensor([1.4, 1.43, 12.2])


class TestOptUtils:
    def test_init_gpr(self):
        likelihood = GaussianLikelihood()
        try:
            myGPR = DragonGPR(X, Y, likelihood=likelihood)
            assert True
        except:
            assert False

    def test_init_gpr_invalid_type(self):
        likelihood = GaussianLikelihood()
        try:
            myGPR = DragonGPR(X, Y, likelihood=likelihood, kernel="RandomKernel")
            assert False
        except:
            assert True

    def test_bopt_init(self):
        # initialize hyperparameters (test fn: lr cannot be a float % 2 == 0)
        def lr_constraint(value: float) -> bool:
            return int(100 * value) % 2 == 0

        lr_param = Hyperparameter(
            name="lr",
            type_="numerical",
            x=0.01,
            constraints=[Constraint(lr_constraint)],
            range_=(1e-10, 1e-2),
        )
        print("Finished lr param init")

        def bad_model(x, lr):
            return 10 * x[0][0]

        # objective function
        def obj(model: callable, input_):
            return model(input_)

        bopt = BaseDragonBayesOpt(
            objective_function=obj,
            Y_init=T.tensor([1.0]),
            hyper_params=[lr_param],
        )

    def test_fail_on_invalid_hyperparam_value(self):
        # initialize hyperparameters (test fn: lr cannot be a float % 2 == 0)
        def lr_constraint(value: float) -> bool:
            return int(100 * value) % 2 == 0

        try:
            lr_param = Hyperparameter(
                name="lr",
                type_="numerical",
                x=0.0,
                constraints=[Constraint(lr_constraint)],
                range_=(1e-10, 1e-2),
            )
        except ValueError:
            return True
        return False

    def test_bopt_sample_from_bounds(self):
        # initialize hyperparameters (test fn: lr cannot be a float % 2 == 0)
        T.manual_seed(0)
        np.random.seed(0)

        def lr_constraint(value: float) -> bool:
            return int(100 * value) % 2 == 0

        lr_param = Hyperparameter(
            name="lr",
            type_="numerical",
            x=0.01,
            constraints=[Constraint(lr_constraint)],
            range_=(1e-10, 1e-2),
        )

        beta0_param = Hyperparameter(
            name="beta0",
            type_="numerical",
            x=0.91,
            range_=(0.9, 0.95),
        )

        beta1_param = Hyperparameter(
            name="beta1",
            type_="numerical",
            x=0.97,
            range_=(0.95, 0.999999),
        )
        print("Finished lr param init")

        def bad_model(x, lr):
            return 10 * x[0][0]

        # objective function
        def obj(model: callable, input_):
            return model(input_)

        bopt = BaseDragonBayesOpt(
            objective_function=obj,
            Y_init=T.tensor([1.0]),
            hyper_params=[lr_param, beta0_param, beta1_param],
        )

        restart_best = bopt._sample_next_points(xi=0.05)
        # check predicted LR

    def test_bopt_sample_from_bounds_expected_improvement(self):
        # initialize hyperparameters (test fn: lr cannot be a float % 2 == 0)
        T.manual_seed(0)
        np.random.seed(0)

        def lr_constraint(value: float) -> bool:
            return int(100 * value) % 2 == 0

        lr_param = Hyperparameter(
            name="lr",
            type_="numerical",
            x=0.01,
            constraints=[Constraint(lr_constraint)],
            range_=(1e-10, 1e-2),
        )

        beta0_param = Hyperparameter(
            name="beta0",
            type_="numerical",
            x=0.91,
            range_=(0.9, 0.95),
        )

        beta1_param = Hyperparameter(
            name="beta1",
            type_="numerical",
            x=0.97,
            range_=(0.95, 0.999999),
        )

        def bad_model(x, lr):
            return 10 * x[0][0]

        # objective function
        def obj(model: callable, input_):
            return model(input_)

        bopt = BaseDragonBayesOpt(
            objective_function=obj,
            Y_init=T.tensor([1.0]),
            hyper_params=[lr_param, beta0_param, beta1_param],
            acquisition_function=expected_improvement,
        )

        restart_best = bopt._sample_next_points(xi=0.05)
        # check predicted LR



import torch as T
import torch.nn
import pytest

from tools.pruning import DistinctivenessPruning
from test_utils.test_models import LeNet, LSTMTS, LSTMTSConfig, device


class TestDistinctivenessPrune:
    def test_get_modules(self):
        # test module retrieval
        model_config = LSTMTSConfig()
        model = LSTMTS(model_config)

        pruner = DistinctivenessPruning(tolerance=(30, 150), model=model, device=device)
        m = pruner.current_modules()
        m1, _ = m["curr"]
        m2, _ = m["next"]
        assert m1 == "lstm" and m2 == "fc2"

    def test_angle_calc_ts(self):
        # test angle calc (result comparison was calculated by hand)
        model_config = LSTMTSConfig()
        model = LSTMTS(model_config)

        # get modules
        pruner = DistinctivenessPruning(tolerance=(30, 150), model=model, device=device)
        m = pruner.current_modules()
        m1, module1 = m["curr"]

        # retrieve weight vector from modules
        param1 = next(module1.parameters())
        wv1 = param1[0]
        wv2 = param1[1]

        result = pruner.get_angle(wv1, wv2)
        assert int(result) == 72  # some wiggle room to allow for fp

    def test_merge_neurons_lenet(self):
        T.manual_seed(0)
        model = LeNet(0.0)
        pruner = DistinctivenessPruning(tolerance=(30, 150), model=model, device=device)
        m = pruner.current_modules()
        m1, module1 = m["curr"]

        # retrieve weight vector from modules
        param1 = next(module1.parameters())
        wv1 = param1[0]
        wv2 = param1[1]
        result1, result2 = pruner.merge_neurons(wv1, wv2)
        assert T.sum(result1).item() - -0.4820 <= 0.0001
        assert T.sum(result2).item() == 0

    def test_prune_parameter_lenet(self):
        T.manual_seed(0)
        model = LeNet(0.0)

        old_weights = model.__getattr__("fc2").weight
        first_sum: float = T.sum(old_weights)

        pruner = DistinctivenessPruning(
            tolerance=(79, 99), model=model, device=device
        )  # Weights mostly orthogonal, average doesn't really change much
        model_, result = pruner(
            "fc2.weight", model, over_next_layer=True, return_result_dict=True
        )

        weights1 = model_.__getattr__("fc2").weight
        weights2 = old_weights
        assert weights1.shape == weights2.shape

        print(T.sum(weights1), first_sum)
        assert not (T.sum(weights1) == first_sum)

    def test_prune_parameter_lstm(self):
        # weights won't be pruned. Assert weights equal
        T.manual_seed(0)
        model_config = LSTMTSConfig()
        model = LSTMTS(model_config)

        old_weights = model.__getattr__("fc2").weight
        first_sum: float = T.sum(old_weights)
        print(first_sum)

        pruner = DistinctivenessPruning(
            tolerance=(79, 99), model=model, device=device
        )  # Weights mostly orthogonal, average doesn't really change much
        model_, result = pruner(
            "fc2.weight", model, over_next_layer=True, return_result_dict=True
        )
        weights1 = model_.__getattr__("fc2").weight
        weights2 = old_weights
        assert weights1.shape == weights2.shape
        assert T.sum(weights1) == first_sum

    def test_invalid_parameter_lenet(self):
        T.manual_seed(0)
        model = LeNet(0.0)

        old_weights = model.__getattr__("conv1").weight
        first_sum: float = T.sum(old_weights)

        pruner = DistinctivenessPruning(tolerance=(79, 99), model=model, device=device)
        try:  # Weights mostly orthogonal, average doesn't really change much
            model_, result = pruner(
                "conv1.weight", model, over_next_layer=True, return_result_dict=True
            )
            assert False

        except ValueError:
            assert True



"""
Testing functions related to the iterative base class pruner
"""

import torch as T
import torch.nn as nn
import pytest

from tools.pruning import IterativeDragonPruner
from test_utils.test_models import LeNet, LSTMTS, LSTMTSConfig, device

# test data

# ground truth LeNet module pair list
module_list = ["conv1", "conv2", "fc1", "fc2", "fc3"]
module_list2 = ["lstm", "fc2"]


class TestDragonPruner:
    # test function for next module
    def test_get_modules(self):
        model = LeNet(0.0)
        pruner = IterativeDragonPruner(model)
        counter = 0
        while pruner.STOP_FLAG == False:
            counter += 1
            pruner._next_module()

            # check pair against ground truth
            m = pruner.current_modules()

            m1, _ = m["curr"]

            if not pruner.STOP_FLAG:
                ground_truth_1 = module_list[counter]

                assert m1 == ground_truth_1

            if counter > 100:
                assert False  # break from test

        assert (
            counter <= 5 and pruner.STOP_FLAG == True
        )  # make sure stop flag and counter work

    # Test apply to for a named param (cnn)
    def test_apply_to_lenet(self):
        model = LeNet(0.0)
        pruner = IterativeDragonPruner(model)

        # define pruning function
        def prune_func(wi: T.tensor):
            return T.zeros_like(wi)

        with T.no_grad():
            result = pruner.apply_to(
                prune_func, model, "fc2.weight", 0, over_next_layer=False
            )

        for result_weight in result[list(result.keys())[0]]:
            key = list(result_weight.keys())[0]
            assert T.sum(result_weight[key][0]) == 0

        for name, module in model.named_parameters():
            if name == "fc2.weight":
                assert T.sum(module[0]) == 0

    # Test apply to for a named param (lstm)
    def test_apply_to_lstm(self):
        model_config = LSTMTSConfig()
        model = LSTMTS(model_config)
        pruner = IterativeDragonPruner(model)

        # define pruning function
        def prune_func(
            wi: T.tensor,
        ):
            return T.zeros_like(wi)

        with T.no_grad():
            result = pruner.apply_to(
                prune_func, model, "lstm.weight_hh_l0", 0, over_next_layer=False
            )

        for result_weight in result[list(result.keys())[0]]:
            key = list(result_weight.keys())[0]
            assert T.sum(result_weight[key][0]) == 0

        for name, module in model.named_parameters():
            if name == "lstm.weight_hh_l0":
                assert T.sum(module[0]) == 0



from setuptools import setup, find_packages

setup(
    name="pytorch-dragon",
    version="0.1.1",  # must match the github version
    author="Campbell Rankine",
    author_email="campbellrankine@gmail.com",
    description="A pytorch integrated Machine Learning / Deep learning utilities library",
    download_url="https://github.com/Campbell-Rankine/dragon/archive/refs/tags/v0.1.0.tar.gz",
    packages=find_packages(),
    install_requires=[  # I get to this in a second
        "validators",
        "beautifulsoup4",
        "opencv-python",
        "numpy",
        "numba",
        "torch==2.2.2",
        "psutil",
        "typing",
        "datetime",
        "gpytorch",
        "botorch",
        "scikit-learn",
        "scipy",
        "torchvision",
        "matplotlib",
        "requests_html",
        "pandas",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.10",
)
